from datetime import datetime

class Competition:
    def __init__(self, initDict):
      parsedDict = {k: parse(v) for k, v in initDict.items()}
      self.__dict__.update(parsedDict)
      self.tags = [Tag(t) for t in self.tags]
    def __repr__(self):
      return self.ref

class SubmitResult:
    def __init__(self, initDict):
      parsedDict = {k: parse(v) for k, v in initDict.items()}
      self.__dict__.update(parsedDict)
    def __repr__(self):
      return self.message

class Submission:
    def __init__(self, initDict):
      parsedDict = {k: parse(v) for k, v in initDict.items()}
      self.__dict__.update(parsedDict)
      self.size = File.getSize(self.totalBytes)
    def __repr__(self):
      return str(self.ref)

class Dataset:
    def __init__(self, initDict):
      parsedDict = {k: parse(v) for k, v in initDict.items()}
      self.__dict__.update(parsedDict)
      self.tags = [Tag(t) for t in self.tags]
      self.size = File.getSize(self.totalBytes)
    def __repr__(self):
      return self.ref

class File:
    def __init__(self, initDict):
      parsedDict = {k: parse(v) for k, v in initDict.items()}
      self.__dict__.update(parsedDict)
      self.size = File.getSize(self.totalBytes)
    def __repr__(self):
      return self.ref
    @staticmethod
    def getSize(size,precision=0):
        suffixes=['B','KB','MB','GB','TB']
        suffixIndex = 0
        while size >= 1024 and suffixIndex < 4:
            suffixIndex += 1
            size = size/1024.0
        return "%.*f%s"%(precision,size,suffixes[suffixIndex])

class Tag:
    def __init__(self, initDict):
      parsedDict = {k: parse(v) for k, v in initDict.items()}
      self.__dict__.update(parsedDict)
    def __repr__(self):
      return self.ref

def parse(string):
  timeFormats = ["%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%dT%H:%M:%S.%f", "%Y-%m-%dT%H:%M:%S.%fZ"]
  for t in timeFormats:
    try:
      result = datetime.strptime(string[:26], t).replace(microsecond=0)
      return result
    except: 
      pass
  return string



from .kaggle_api import KaggleApi
from ..models.kaggle_models_extended import Competition, SubmitResult, Submission, Dataset, File
from kaggle.configuration import Configuration
import os, json, sys, csv, zipfile
from os.path import expanduser, isfile
from datetime import datetime

class KaggleApi(KaggleApi):
    configPath = os.path.join(expanduser("~"),".kaggle")
    configFile = 'kaggle.json'
    config = os.path.join(configPath, configFile)
    path = os.path.join(expanduser("~"),".kaggle")

    def authenticate(self):
      try:
        configuration = Configuration()
        if os.name != 'nt':
          permissions = os.stat(self.config).st_mode
          if (permissions & 4) or (permissions & 32):
            print("Warning: Your Kaggle API key is readable by other users on this system! To fix this, you can run 'chmod 600 {}'".format(self.config))
        with open(self.config, 'r') as f:
          configData = json.load(f)
        configuration.username = configData['username']
        configuration.password = configData['key']
        self.api_client.configuration = configuration
      except: 
        sys.exit('Unauthorized: you must download an API key from https://www.kaggle.com/<username>/account\nThen put ' + self.configFile + ' in the folder ' + self.configPath)

    def downloadPath(self, path = None, quiet = False):
      try: 
        os.makedirs(os.path.dirname(self.configPath), exist_ok = True)
        with open(self.config, 'r') as f:
          configData = json.load(f)
        if path is not None:
          configData['path'] = path
          with open(self.config, 'w') as f:
            json.dump(configData, f)
        if 'path' in configData:
          self.path = configData['path']
      except:
        pass
      if not quiet:
        print('Your files will be downloaded to ' + self.path)
   
    def competitionsList(self, page = 1, search = ''):
      if search is None:
        search = ''
      competitionsListResult = self.competitions_list(page = page, search = search)
      return [Competition(c) for c in competitionsListResult]

    def competitionsListCli(self, page = 1, search = '', csv = False):
      competitions = self.competitionsList(page, search)
      fields = ['ref', 'deadline', 'category', 'reward', 'teamCount', 'userHasEntered']
      if competitions:
        if csv:
           self.printCsv(competitions, fields)
        else:
           self.printTable(competitions, fields)
      else:
        print('No competitions found')

    def competitionSubmit(self, file, message, competition):
      urlResult = self.competitions_submissions_url(file_name = os.path.basename(file), content_length = os.path.getsize(file), last_modified_date_utc = int(os.path.getmtime(file) * 1000))
      urlResultList = urlResult['createUrl'].split('/')
      uploadResult = self.competitions_submissions_upload(file = file, guid = urlResultList[-3], content_length = urlResultList[-2], last_modified_date_utc = urlResultList[-1])
      uploadResultToken = uploadResult['token']
      submitResult = self.competitions_submissions_submit(id = competition, blob_file_tokens = uploadResultToken, submission_description = message)
      return SubmitResult(submitResult)

    def competitionSubmissions(self, competition):
      submissionsResult = self.competitions_submissions_list(id = competition)
      return [Submission(s) for s in submissionsResult]

    
    def competitionSubmissionsCli(self, competition, csv = False):
      submissions = self.competitionSubmissions(competition)
      fields = ['fileName', 'date', 'description', 'status', 'publicScore', 'privateScore']
      if submissions:
        if csv:
          self.printCsv(submissions, fields)
        else:
          self.printTable(submissions, fields)
      else:
        print('No submissions found')

    def competitionListFiles(self, competition):
      competitionListFilesResult = self.competitions_data_list_files(id = competition)
      return [File(f) for f in competitionListFilesResult]

    def competitionListFilesCli(self, competition, csv = False):
      files = self.competitionListFiles(competition)
      fields = ['name', 'size', 'creationDate']
      if files:
        if csv:
          self.printCsv(fiels, fields)
        else:
          self.printTable(files, fields)
      else:
        print('No files found')

    def competitionDownloadFile(self, competition, file, path = None, force = False, quiet = True):
      if path is None:
        path = os.path.join(self.path, 'competitions')
      response = self.competitions_data_download_file(id = competition, file_name = file, _preload_content = False)
      url = response.retries.history[0].redirect_location.split('?')[0]
      outfile = os.path.join(path, competition, url.split('/')[-1])
      if force or self.downloadNeeded(response, outfile, quiet):
        self.downloadFile(response, outfile, quiet)

    def competitionDownloadFiles(self, competition, path = None, force = False, quiet = True):
      files = self.competitionListFiles(competition)
      for file in files:
        self.competitionDownloadFile(competition, file.ref, path, force, quiet)

    def competitionDownloadCli(self, competition, file = None, path = None, force = False, quiet = False):
      if file is None:
        self.competitionDownloadFiles(competition, path, force, quiet)
      else:
        self.competitionDownloadFile(competition, file, path, force, quiet)

    def datasetsList(self, page = 1, search = ''):
      if search is None:
        search = ''
      datasetsListResult = self.datasets_list(page = page, search = search)
      return [Dataset(d) for d in datasetsListResult]

    def datasetsListCli(self, page = 1, search = '', csv = False):
      datasets = self.datasetsList(page, search)
      fields = ['ref', 'title', 'size', 'lastUpdated', 'downloadCount']
      if datasets:
        if csv:
          self.printCsv(datasets, fields)
        else:
          self.printTable(datasets, fields)
      else:
        print('No datasets found')

    def datasetListFiles(self, dataset):
      datasetUrlList = dataset.split('/')
      ownerSlug = datasetUrlList[0]
      datasetSlug = datasetUrlList[1]
      datasetListFilesResult = self.datasets_list_files(owner_slug = ownerSlug, dataset_slug = datasetSlug)
      return [File(f) for f in datasetListFilesResult]

    def datasetListFilesCli(self, dataset, csv = False):
      files = self.datasetListFiles(dataset)
      fields = ['name', 'size', 'creationDate']
      if files:
        if csv:
          self.printCsv(files, fields)
        else:
          self.printTable(files, fields)
      else:
        print('No files found')

    def datasetDownloadFile(self, dataset, file, path = None, force = False, quiet = True):
      if path is None:
        path = os.path.join(self.path, 'datasets')
      datasetUrlList = dataset.split('/')
      ownerSlug = datasetUrlList[0]
      datasetSlug = datasetUrlList[1]
      response = self.datasets_download_file(owner_slug = ownerSlug, dataset_slug = datasetSlug, file_name = file, _preload_content = False)
      url = response.retries.history[0].redirect_location.split('?')[0]
      outfile = os.path.join(path, ownerSlug, datasetSlug, url.split('/')[-1])
      if force or self.downloadNeeded(response, outfile, quiet):
        self.downloadFile(response, outfile, quiet)

    def datasetDownloadFiles(self, dataset, path = None, force = False, quiet = True):
      if path is None:
        path = os.path.join(self.path, 'datasets')
      datasetUrlList = dataset.split('/')
      ownerSlug = datasetUrlList[0]
      datasetSlug = datasetUrlList[1]
      self.datasetDownloadFile(dataset, datasetSlug + '.zip', path, force, quiet)
      outpath = os.path.join(path, ownerSlug, datasetSlug)
      outfile = os.path.join(outpath, datasetSlug + '.zip')
      with zipfile.ZipFile(outfile, 'r') as z:
        z.extractall(outpath)

    def datasetDownloadCli(self, dataset, file = None, path = None, force = False, quiet = False):
      if file is None:
        self.datasetDownloadFiles(dataset, path, force, quiet)
      else:
        self.datasetDownloadFile(dataset, file, path, force, quiet)

    def downloadFile(self, response, outfile, quiet = True, chunkSize = 1048576):
      os.makedirs(os.path.dirname(outfile), exist_ok = True)
      size = int(response.headers['Content-Length'])
      sizeRead = 0
      with open(outfile, 'wb') as out:
        while True:
          data = response.read(chunkSize)
          if not data:
            break
          out.write(data)
          sizeRead = min(size, sizeRead + chunkSize) 
          if not quiet:
            print(os.path.basename(outfile) + ': Downloaded ' + File.getSize(sizeRead) + ' of ' + File.getSize(size), end = '\r')
      if not quiet:
        print('\n', end = '')

    def downloadNeeded(self, response, outfile, quiet = True):
      try:
        remoteDate = datetime.strptime(response.headers['Last-Modified'], "%a, %d %b %Y %X %Z")
        if isfile(outfile):
          localDate = datetime.fromtimestamp(os.path.getmtime(outfile))
          if remoteDate <= localDate:
            if not quiet:
              print(os.path.basename(outfile) + ': Skipping, found more recently modified local copy (use --force to force download)')
            return False
      except:
         pass
      return True

    def printTable(self, items, fields):
      formats = []
      borders = []
      for f in fields:
        length = max(len(f),max([len(str(getattr(i,f))) for i in items]))
        justify = '>' if isinstance(getattr(items[0],f), int) or f == 'size' or f == 'reward' else '<'
        formats.append('{:' + justify + str(length + 2) + '}')
        borders.append('-' * length + '  ')
      row_format = ''.join(formats)
      headers = [f + '  ' for f in fields]
      print(row_format.format(*headers))
      print(row_format.format(*borders))
      for i in items:
        i_fields = [str(getattr(i,f)) + '  ' for f in fields]
        print(row_format.format(*i_fields))

    def printCsv(self, items, fields):
      writer = csv.writer(sys.stdout)
      writer.writerow(fields)
      for i in items:
        i_fields = [str(getattr(i,f)) for f in fields]
        writer.writerow(i_fields)



import argparse
from kaggle import api

def main():
  parser = argparse.ArgumentParser(formatter_class = argparse.RawTextHelpFormatter)
  subparsers = parser.add_subparsers(title = 'commands', help = Help.kaggle, dest = 'command')
  subparsers.required = True
  subparsers.choices = Help.kaggle_choices
  parse_competitions(subparsers)
  parse_datasets(subparsers)
  parse_config(subparsers)
  args = parser.parse_args()
  command_args = {}
  command_args.update(vars(args))
  del command_args['func']
  del command_args['command']
  out = args.func(**command_args)
  if out is not None:
    print(out,end='')

def parse_competitions(subparsers):
  parser_competitions = subparsers.add_parser('competitions', formatter_class = argparse.RawTextHelpFormatter)
  subparsers_competitions = parser_competitions.add_subparsers(title = 'commands', dest = 'command')
  subparsers_competitions.required = True
  subparsers_competitions.choices = Help.competitions_choices

  parser_competitions_list = subparsers_competitions.add_parser('list', formatter_class = argparse.RawTextHelpFormatter)
  parser_competitions_list_optional = parser_competitions_list._action_groups.pop()
  parser_competitions_list_optional.add_argument('-p', '--page', dest = 'page', default = 1, required = False, help = Help.page)
  parser_competitions_list_optional.add_argument('-s', '--search', dest = 'search', required = False, help = Help.search)
  parser_competitions_list_optional.add_argument('-v', '--csv', dest = 'csv', action = 'store_true', help = Help.csv)
  parser_competitions_list._action_groups.append(parser_competitions_list_optional)
  parser_competitions_list.set_defaults(func = api.competitionsListCli)

  parser_competitions_submit = subparsers_competitions.add_parser('submit', formatter_class = argparse.RawTextHelpFormatter)
  parser_competitions_submit_optional = parser_competitions_submit._action_groups.pop()
  parser_competitions_submit_required = parser_competitions_submit.add_argument_group('required arguments')
  parser_competitions_submit_required.add_argument('-c', '--competition', dest = 'competition', required = True, help = Help.competition)
  parser_competitions_submit_required.add_argument('-f', '--file', dest = 'file', required = True, help = Help.upfile)
  parser_competitions_submit_required.add_argument('-m', '--message', dest = 'message', required = True, help = Help.message)
  parser_competitions_submit._action_groups.append(parser_competitions_submit_optional)
  parser_competitions_submit.set_defaults(func = api.competitionSubmit)

  parser_competitions_submissions = subparsers_competitions.add_parser('submissions', formatter_class = argparse.RawTextHelpFormatter)
  parser_competitions_submissions_optional =  parser_competitions_submissions._action_groups.pop()
  parser_competitions_submissions_required = parser_competitions_submissions.add_argument_group('required arguments')
  parser_competitions_submissions_required.add_argument('-c', '--competition', dest = 'competition', required = True, help = Help.competition)
  parser_competitions_submissions_optional.add_argument('-v', '--csv', dest = 'csv', action = 'store_true', help = Help.csv)
  parser_competitions_submissions._action_groups.append(parser_competitions_submissions_optional)
  parser_competitions_submissions.set_defaults(func = api.competitionSubmissionsCli)  

  parser_competitions_files = subparsers_competitions.add_parser('files', formatter_class = argparse.RawTextHelpFormatter)
  parser_competitions_files_optional =  parser_competitions_files._action_groups.pop()
  parser_competitions_files_required = parser_competitions_files.add_argument_group('required arguments')
  parser_competitions_files_required.add_argument('-c', '--competition', dest = 'competition', required = True, help = Help.competition)
  parser_competitions_files_optional.add_argument('-v', '--csv', dest = 'csv', action = 'store_true', help = Help.csv)
  parser_competitions_files._action_groups.append(parser_competitions_files_optional)
  parser_competitions_files.set_defaults(func = api.competitionListFilesCli)  

  parser_competitions_download = subparsers_competitions.add_parser('download', formatter_class = argparse.RawTextHelpFormatter)
  parser_competitions_download_optional = parser_competitions_download._action_groups.pop()
  parser_competitions_download_required = parser_competitions_download.add_argument_group('required arguments')
  parser_competitions_download_required.add_argument('-c', '--competition', dest = 'competition', required = True, help = Help.competition)
  parser_competitions_download_optional.add_argument('-f', '--file', dest = 'file', required = False, help = Help.competition_file)
  parser_competitions_download_optional.add_argument('-p', '--path', dest = 'path', required = False, help = Help.path)
  parser_competitions_download_optional.add_argument('-w', '--wp', dest = 'path', action = 'store_const', const = '.', required = False, help = Help.wp)
  parser_competitions_download_optional.add_argument('-o', '--force', dest = 'force', action = 'store_true', help = Help.force)
  parser_competitions_download_optional.add_argument('-q', '--quiet', dest = 'quiet', action = 'store_true', help = Help.quiet)
  parser_competitions_download._action_groups.append(parser_competitions_download_optional)
  parser_competitions_download.set_defaults(func = api.competitionDownloadCli)  

def parse_datasets(subparsers):
  parser_datasets = subparsers.add_parser('datasets', formatter_class = argparse.RawTextHelpFormatter)
  subparsers_datasets = parser_datasets.add_subparsers(title = 'commands', dest = 'command')
  subparsers_datasets.required = True
  subparsers_datasets.choices = Help.datasets_choices

  parser_datasets_list = subparsers_datasets.add_parser('list', formatter_class = argparse.RawTextHelpFormatter)
  parser_datasets_list_optional = parser_datasets_list._action_groups.pop()
  parser_datasets_list.add_argument('-p', '--page', dest = 'page', default = 1, required = False, help = Help.page)
  parser_datasets_list.add_argument('-s', '--search', dest = 'search', required = False, help = Help.search)
  parser_datasets_list.add_argument('-v', '--csv', dest = 'csv', action = 'store_true', help = Help.csv)
  parser_datasets_list._action_groups.append(parser_datasets_list_optional)
  parser_datasets_list.set_defaults(func = api.datasetsListCli)

  parser_datasets_files = subparsers_datasets.add_parser('files', formatter_class = argparse.RawTextHelpFormatter)
  parser_datasets_files_optional = parser_datasets_files._action_groups.pop()
  parser_datasets_files_required = parser_datasets_files.add_argument_group('required arguments')
  parser_datasets_files_required.add_argument('-d', '--dataset', dest = 'dataset', required = True, help = Help.dataset)
  parser_datasets_files_optional.add_argument('-v', '--csv', dest = 'csv', action = 'store_true', help = Help.csv)
  parser_datasets_files._action_groups.append(parser_datasets_files_optional)
  parser_datasets_files.set_defaults(func = api.datasetListFilesCli)  

  parser_datasets_download = subparsers_datasets.add_parser('download', formatter_class = argparse.RawTextHelpFormatter)
  parser_datasets_download_optional = parser_datasets_download._action_groups.pop()
  parser_datasets_download_required = parser_datasets_download.add_argument_group('required arguments')
  parser_datasets_download_required.add_argument('-d', '--dataset', dest = 'dataset', required = True, help = Help.dataset)
  parser_datasets_download_optional.add_argument('-f', '--file', dest = 'file', required = False, help = Help.dataset_file)
  parser_datasets_download_optional.add_argument('-p', '--path', dest = 'path', required = False, help = Help.path)
  parser_datasets_download_optional.add_argument('-w', '--wp', dest = 'path', action = 'store_const', const = '.', required = False, help = Help.wp)
  parser_datasets_download_optional.add_argument('-o', '--force', dest = 'force', action = 'store_true', help = Help.force)
  parser_datasets_download_optional.add_argument('-q', '--quiet', dest = 'quiet', action = 'store_true', help = Help.quiet)
  parser_datasets_download._action_groups.append(parser_datasets_download_optional)
  parser_datasets_download.set_defaults(func = api.datasetDownloadCli)  

def parse_config(subparsers):
  parser_config = subparsers.add_parser('config', formatter_class = argparse.RawTextHelpFormatter)
  subparsers_config = parser_config.add_subparsers(title = 'commands', dest = 'command')
  subparsers_config.required = True
  subparsers_config.choices = Help.config_choices

  parser_config_path = subparsers_config.add_parser('path', formatter_class = argparse.RawTextHelpFormatter, help = Help.download_path)
  parser_config_path_optional = parser_config_path._action_groups.pop()
  parser_config_path_optional.add_argument('-p', '--path', dest = 'path', required = False, help = Help.path)
  parser_config_path._action_groups.append(parser_config_path_optional)
  parser_config_path.set_defaults(func = api.downloadPath)  

class Help:  
  kaggle_choices = ['competitions', 'datasets', 'config']
  competitions_choices = ['list', 'submit', 'submissions', 'files', 'download']
  datasets_choices = ['list', 'files', 'download']
  config_choices = ['path']

  kaggle = 'Use one of:\ncompetitions {' + ', '.join(competitions_choices) +'}\ndatasets {' + ', '.join(datasets_choices) + '}\nconfig {' + ', '.join(config_choices) + '}'
  competition = 'competition URL suffix\n(use "kaggle competitions list" to show options)'
  path = 'folder where file(s) will be downloaded, defaults to ' + api.configPath
  force = 'skip check whether local version of file is up to date, force file download'
  quiet = 'suppress printing information about download progress'
  dataset = 'dataset URL suffix in format <owner>/<dataset-name>\n(use "kaggle datasets list" to show options)'
  page = 'page number'
  upfile = 'file for upload, including path'
  message = 'message describing this submission'
  competition_file = 'file name, all files downloaded if not provided\n(use "kaggle competitions files -c <competition>" to show options)'
  dataset_file = 'file name, all files downloaded if not provided\n(use "kaggle datasets files -d <dataset>" to show options)'
  download_path = 'check or set folder where files will be downloaded'
  csv = 'print in CSV format\n(if not set print in table format)'
  search = 'term(s) to search for'
  wp = 'download to current working path'


from __future__ import absolute_import
from kaggle.api.kaggle_api_extended import KaggleApi
from kaggle.api_client import ApiClient

api = KaggleApi(ApiClient())
api.authenticate()
api.downloadPath(quiet = True)
